from zope.interface import Interface
from zope.schema import TextLine
from zope.configuration.fields import GlobalObject, Path

from hurry import custom

class ITemplateLanguage(Interface):
    template_class = GlobalObject(
        title=u'Template Class',
        description=u'The class that implements the template language')

    extension = TextLine(
        title=u'Extension',
        description=u'The filesystem extension (with period, '
                    u'for example .jsont) '
                    u'recognized for this template language')

    sample_extension = TextLine(
        title=u'Sample Extension',
        description=u'The filesystem extension used for sample input data.',
        required=False)

class IDataLanguage(Interface):
    parse_func = GlobalObject(
        title=u'Parse function',
        description=u'A function that can parse text and deliver input data.')

    extension = TextLine(
        title=u'Extension',
        description=u'The filesystem extension (with period, '
                    u'for example .json) '
                    u'recognzied for this input language')

class ICollection(Interface):
    id = TextLine(
        title=u'Collection ID',
        description=u"Globally unique collection id")

    path = Path(
        title=u"Path",
        description=u'Filesystem path to collection')

    title = TextLine(
        title=u"Title",
        description=u'Human-readable title for collection',
        required=False)

def action_template_language(_context,
                             template_class, extension, sample_extension=None):
    _context.action(
        discriminator = ('action_template_language',
                         template_class, extension, sample_extension),
        callable = custom.register_language,
        args = (template_class, extension, sample_extension))
    
def action_data_language(_context, parser_func, extension):
    _context.action(
        discriminator = ('action_data_language',
                         parser_func, extension),
        callable = custom.register_data_language,
        args = (parser_func, extension))
    
def action_collection(_context, id, path, title=None):
    _context.action(
        discriminator = ('action_collection', id),
        callable = custom.register_collection,
        args = (id, path, title))



import jsontemplate
from zope.interface import implements
from hurry.custom.interfaces import ITemplate, CompileError, RenderError

class JsonTemplate(object):
    implements(ITemplate)

    def __init__(self, source):
        try:
            self.json_template = jsontemplate.Template(source)
        except jsontemplate.CompilationError, e:
            raise CompileError(unicode(e))
        self.source = source
        
    def __call__(self, input):
        try:
            return self.json_template.expand(input)
        except jsontemplate.EvaluationError, e:
            raise RenderError(unicode(e))

        



from zope.interface import moduleProvides

from hurry.custom.core import (render,
                               lookup,
                               check,
                               structure,
                               collection,
                               root_collection,
                               next_collection,
                               register_language,
                               register_data_language,
                               register_collection,
                               recognized_languages)

from hurry.custom.core import (FilesystemTemplateDatabase,
                               InMemoryTemplateDatabase)

from hurry.custom.interfaces import IHurryCustomAPI

from hurry.custom.jsont import JsonTemplate

moduleProvides(IHurryCustomAPI)
__all__ = list(IHurryCustomAPI)



from zope.testing.cleanup import addCleanUp
from zope import component
from zope.component import registry

# a very simple implementation of setSite and getSite so we don't have
# to rely on zope.app.component just for our tests
_site = None

class DummySite(object):
    def __init__(self, id):
        self.id = id
        self._sm = SiteManager()
        
    def getSiteManager(self):
        return self._sm

class SiteManager(registry.Components):
    def __init__(self):
        super(SiteManager, self).__init__()
        self.__bases__ = (component.getGlobalSiteManager(),)

def setSite(site=None):
    global _site
    _site = site

def getSite():
    return _site

def adapter_hook(interface, object, name='', default=None):
    try:
        return getSiteManager().adapters.adapter_hook(
            interface, object, name, default)
    except component.interfaces.ComponentLookupError:
        return default

def getSiteManager(context=None):
    if _site is not None:
        return _site.getSiteManager()
    return component.getGlobalSiteManager()

def setHooks():
    component.adapter_hook.sethook(adapter_hook)
    component.getSiteManager.sethook(getSiteManager)

def resetHooks():
    component.adapter_hook.reset()
    component.getSiteManager.reset()

# make sure hooks get cleaned up after tests are run
addCleanUp(resetHooks)



import os, time, glob
from datetime import datetime
from zope.interface import implements
from zope import component
from zope.component.interfaces import ComponentLookupError
from hurry.custom.interfaces import (
    ITemplate, IManagedTemplate, ITemplateDatabase, IDataLanguage,
    ISampleExtension, CompileError, RenderError, NotSupported)

def register_language(template_class, extension, sample_extension=None):
    component.provideUtility(template_class,
                             provides=ITemplate,
                             name=extension)
    if sample_extension is not None:
        component.provideUtility(sample_extension,
                                 provides=ISampleExtension,
                                 name=extension)

def register_data_language(parse_func, extension):
    component.provideUtility(parse_func,
                             provides=IDataLanguage,
                             name=extension)

def recognized_languages():
    return component.getUtilitiesFor(ITemplate)

def register_collection(id, path, title=None):
    if title is None:
        title = id
    db = FilesystemTemplateDatabase(id=id, path=path, title=title)
    component.provideUtility(db,
                             provides=ITemplateDatabase,
                             name=id)

def render(id, template_path, input):
    template = lookup(id, template_path)
    while True:
        try:
            return template(input)
        except RenderError, render_error:
            try:
                next_db = next_collection(id, template.db)
            except ComponentLookupError:
                # cannot find any next collection, so this error is it
                raise render_error
            template = lookup(id, template_path, db=next_db)

def lookup(id, template_path, db=None):
    dummy, ext = os.path.splitext(template_path)
    template_class = component.getUtility(ITemplate, name=ext)

    db = db or collection(id)

    while True:
        source = db.get_source(template_path)
        if source is not None:
            try:
                return ManagedTemplate(template_class, db, template_path)
            except CompileError, e:
                pass
        try:
            db = next_collection(id, db)
        except ComponentLookupError:
            # if we cannot find a next collection, this means the
            # last collection had a fatal compilation error
            raise e

def collection(id):
    return component.getUtility(ITemplateDatabase, name=id)

def next_collection(id, db):
    result = getNextUtility(db, ITemplateDatabase, name=id)
    if result is db:
        raise ComponentLookupError("No collection available for: %s" % id)
    return result

def root_collection(id):
    db = collection(id)
    while True:
        try:
            next_db = next_collection(id, db)
        except ComponentLookupError:
            return db
        db = next_db

def check(id, template_path, source):
    dummy, ext = os.path.splitext(template_path)
    template_class = component.getUtility(ITemplate, name=ext)
    # can raise CompileError
    template = template_class(source)
    db = root_collection(id)
    samples = db.get_samples(template_path)
    for key, value in samples.items():
        try:
            template(value)
        except RenderError, e:
            # add data_id and re-raise
            e.data_id = key
            raise e

def structure(id):
    extensions = set([extension for
                      (extension, language) in recognized_languages()])
    db = root_collection(id)
    return _get_structure_helper(db.path, db.path, extensions)

class ManagedTemplate(object):
    implements(IManagedTemplate)
    
    def __init__(self, template_class, db, template_path):
        self.template_class = template_class
        self.db = db
        self.template_path = template_path
        self.load()
        self._last_updated = 0
        
    def load(self):
        self.template = self.template_class(
            self.db.get_source(self.template_path))

    def check(self):
        mtime = self.db.get_modification_time(self.template_path)
        if mtime > self._last_updated:
            self._last_updated = mtime
            self.load()

    @property
    def source(self):
        self.check()
        return self.template.source

    def __call__(self, input):
        self.check()
        return self.template(input)
    
class FilesystemTemplateDatabase(object):
    implements(ITemplateDatabase)

    template_encoding = 'UTF-8'
    
    def __init__(self, id, path, title):
        self.id = id
        self.path = path
        self.title = title

    def update(self, template_id, source):
        raise NotSupported(
            "Cannot update templates in FilesystemTemplateDatabase.")

    def get_source(self, template_id):
        template_path = os.path.join(self.path, template_id)
        f = open(template_path, 'r')
        result = f.read()
        f.close()
        return unicode(result, self.template_encoding)
    
    def get_modification_time(self, template_id):
        template_path = os.path.join(self.path, template_id)
        return os.path.getmtime(template_path)

    def get_samples(self, template_id):
        template_path = os.path.join(self.path, template_id)
        template_dir = os.path.dirname(template_path)
        template_name, extension = os.path.splitext(template_id)
        result = {}
        sample_extension = component.queryUtility(ISampleExtension,
                                                  name=extension,
                                                  default=None)
        if sample_extension is None:
            return result
        parse = component.getUtility(IDataLanguage, name=sample_extension)
        for path in glob.glob(
            os.path.join(template_dir,
                         template_name + '-*' + sample_extension)):
            filename = os.path.basename(path)
            name, dummy = os.path.splitext(filename)
            # +1 to adjust for -
            name = name[len(template_name) + 1:]        
            f = open(path, 'rb')
            data = f.read()
            f.close()
            result[name] = parse(data)
        return result

class InMemoryTemplateSource(object):
    def __init__(self, source):
        self.source = source
        self.last_updated = time.time()

class InMemoryTemplateDatabase(object):
    implements(ITemplateDatabase)
    
    def __init__(self, id, title):
        self.id = id
        self.title = title
        self._templates = {}

    def update(self, template_id, source):
        self._templates[template_id] = InMemoryTemplateSource(source)

    def get_source(self, template_id):
        try:
            return self._templates[template_id].source
        except KeyError:
            return None
        
    def get_modification_time(self, template_id):
        try:
            return self._templates[template_id].last_updated
        except KeyError:
            return None

    def get_samples(self, template_id):
        return {}

def _get_structure_helper(path, collection_path, extensions):
    entries = os.listdir(path)
    result = []
    for entry in entries:
        entry_path = os.path.join(path, entry)
        if os.path.isdir(entry_path):
            info = {
                'directory': entry,
                'entries': _get_structure_helper(entry_path,
                                                 collection_path, extensions),
                'path': relpath(entry_path, collection_path),
                }
            result.append(info)
        else:
            name, ext = os.path.splitext(entry)
            if ext not in extensions:
                continue
            info = {
                'template': entry,
                'name': name,
                'extension': ext,
                'path': relpath(entry_path, collection_path),
                }
            result.append(info)
    return result

# XXX copied from zope.app.component to avoid dependency on it
# note that newer versions of zope.component have this, so
# when the target app depends on that we can switch and
# eliminate this code

from zope.component import getSiteManager

_marker = object()

def queryNextUtility(context, interface, name='', default=None):
    """Query for the next available utility.

    Find the next available utility providing `interface` and having the
    specified name. If no utility was found, return the specified `default`
    value.
    """
    sm = getSiteManager(context)
    bases = sm.__bases__
    for base in bases:
        util = base.queryUtility(interface, name, _marker)
        if util is not _marker:
            return util
    return default

def getNextUtility(context, interface, name=''):
    """Get the next available utility.

    If no utility was found, a `ComponentLookupError` is raised.
    """
    util = queryNextUtility(context, interface, name, _marker)
    if util is _marker:
        raise ComponentLookupError(
            "No more utilities for %s, '%s' have been found." % (
                interface, name))
    return util

# XXX this code comes from Python 2.6 - when switching to this
# python version we can import it from os.path and get rid of this code
from os.path import commonprefix, abspath, join, sep, pardir

def relpath(path, start):
    """Return a relative version of a path"""

    if not path:
        raise ValueError("no path specified")

    start_list = abspath(start).split(sep)
    path_list = abspath(path).split(sep)

    # Work out how much of the filepath is shared by start and path.
    i = len(commonprefix([start_list, path_list]))

    rel_list = [pardir] * (len(start_list)-i) + path_list[i:]
    return join(*rel_list)



import unittest
import doctest

from hurry import custom

from zope.testing import cleanup
import zope.component.eventtesting
from zope.configuration.xmlconfig import XMLConfig

from hurry.custom.testing import setHooks, setSite, getSite, DummySite

def setUpCustom(test):
    # set up special local component architecture
    setHooks()
    
    # set up event handling
    zope.component.eventtesting.setUp(test)

    # set up the directives
    XMLConfig('meta.zcml', custom)()

def tearDownCustom(test):
    # clean up Zope
    cleanup.cleanUp()
    setSite(None)

def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    globs = {
        'DummySite': DummySite,
        'setSite': setSite,
        'getSite': getSite,
        }
    
    suite = unittest.TestSuite()
    
    suite.addTest(doctest.DocFileSuite(
        'README.txt', 'jsontemplate.txt', 'zcml.txt',
        optionflags=optionflags,
        setUp=setUpCustom,
        tearDown=tearDownCustom,
        globs=globs))
    
    return suite



from setuptools import setup, find_packages
import os

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

long_description = (
    read('src', 'hurry', 'custom', 'README.txt')
    + '\n' +
    read('CHANGES.txt')
    + '\n' +
    'Download\n'
    '========\n'
    )

setup(
    name="hurry.custom",
    version = '0.6',
    description="A framework for allowing customizing templates",
    long_description=long_description,    
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    keywords='template custom templating customization',
    author='Martijn Faassen, Startifact',
    author_email='faassen@startifact.com',
    url='',
    license='ZPL 2.1',
    packages=find_packages('src'),
    package_dir= {'':'src'},
    namespace_packages=['hurry'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
       'setuptools',
       'zope.component',
       'zope.interface',
       'zope.hookable',
       'jsontemplate',
       'zope.configuration',
       ],
    )



from zope.interface import Interface, Attribute

class IHurryCustomAPI(Interface):
    """API for hurry.custom.
    """
    def register_language(template_class, extension, sample_extension=None):
        """Register a template language with the system.

        template_class - a class which implements ITemplate
        extension - the filename extension to register this
                    language. (example: .jsont)
        sample_extension - the filename extension of sample data files
                           for an extension. (example: .json)
        """

    def register_data_language(parse_func, extension):
        """Register a data language for template input.

        parse_func - a function that takes a text and parses it into
                     a data structure.
        extension - the extension to register the data language under.
                     (example: .json).
        """
        
    def register_collection(id, path, title=None):
        """Register a collection of templates on the filesystem with the system.

        id - globally unique collection id (used to look up the collection)
        path - the path of the collection on the filesystem
        title - optionally, human-readable title for collection.
                By default the 'id' will be used.
        """

    def render(id, template_path, input):
        """Render a template.

        id - the id for the collection
        template_path - the relative path (or filename) of the template
                        itself, under the path of the collection
        input - input data for the template

        If the template raises a CompileError or RenderError, the
        system will automatically fall back on the original
        non-customized template.
        """
        
    def lookup(id, template_path):
        """Look up template.
        
        id - the id of the collection
        template_path - the relative path (or filename) of the template
                        itself, under the path of the collection
        """

    def collection(id):
        """Look up ITemplateDatabase for id.

        id - the id of the collection

        Will raise ComponentLookupError if no collection with this id exists.
        """

    def next_collection(id, db):
        """Look up the collection below this one.

        id - the id of the collection
        db - the db below which to look.

        Will raise ComponentLookupError if there is no db below this
        collection.

        Will also raise ComponentLookupError if no collection with this id exists.
        """

    def root_collection(id):
        """Look up the root collection.

        id - the id of the collection.

        Will raise ComponentLookupError if no collection with this id exists.
        """
        
    def check(id, template_path, source):
        """Test a template (before customization).

        id - the id for the collection
        template_path - the template path of the template being customized
        source - the source of the customized template

        This tries a test-compile of the template. If the
        compilation cannot proceed, a CompileError is raised.

        Then tries to render the template with any sample inputs.
        If a rendering fails, a RenderError is raised. In a special
        'data_id' attribute of the error the failing input data is
        indicated.

        If the check succeeds, no exception is raised.
        """
        
    def structure(id):
        """Get a list with all the templates in this collection.

        id - the collection id
        
        All recognized template extensions are reported; unrecognized
        extensions are ignored. Subdirectories are also reported.
    
        Returned is a list of all entries.
    
        List entries for templates look like this:
        
        { 'template': 'template1.st',
        'name': 'template1',
        'extension': '.st'
        'path': 'template1.st',
        }

        template: the name of the template as it is in its immediate
                  directory.
        name: the name of the template without extension
        extension: the template extension
        path: the relative path to the extension from the collection_path
    
        List entries for sub directories look like this:
    
        { 'directory': 'subdirname',
        'entries': [ ... ],
        'path': 'subdirname',
        }

        directory: the name of the directory
        entries: the entries of the subdirectory, in a list
        path: the relative path to this directory from the collection_path
        """

    def recognized_languages():
        """Get an iterable with the recognized languages.
        
        The items are name-value pairs (language extension, template class).
        """
    
class ITemplate(Interface):
    """Uses for classes implementing a template language.

    When creating an object that provides ITemplate, raise
    a CompileError if the template text cannot be processed.
    """
    source = Attribute("The source text of the template.")

    def __call__(input):
        """Render the template given input.

        input - opaque template-language native data structure.

        Raise a RenderError if the template cannot be rendered.
        """

class CompileError(Exception):
    """Error when a template is broken (cannot be parsed/compiled).
    """

class RenderError(Exception):
    """Error when an error cannot be rendered (incorrect input data or
    other run-time error.
    """

class IDataLanguage(Interface):
    def __call__(data):
        """Parse data into data structure that can be passed to ITemplate()"""

class ISampleExtension(Interface):
    """Marker interface used to register the extension of the sample language.
    """

class IManagedTemplate(ITemplate):
    template = Attribute("The real template object being managed.")

    def check():
        """Update the template if it has changed.
        """

    def load():
        """Load the template from the filesystem.
        """

class NotSupported(Exception):
    pass

class ITemplateDatabase(Interface):
    """A per-collection template database.
    """
    id = Attribute("The id of the collection")
    title = Attribute("The title of the collection")

    def update(template_id, source):
        """Update the source for a given template.

        Updates the source and modification time for the template with
        that id.

        If this operation is not supported, a NotSupported error is
        raised.
        """

    def get_source(template_id):
        """Get the source of a given template.

        Returns None if the source cannot be loaded.
        """

    def get_modification_time(template_id):
        """Get the time at which a template was last updated.

        Time must be in number of seconds since epoch (preferably with
        sub-second accuracy, but this is database dependent).

        Returns None if the time cannot be retrieved.
        """
        
    def get_samples(template_id):
        """Get samples for a given template.

        Returns a dictionary with sample inputs.

        keys are the unique ids for the sample inputs.
        values are the actual template-language native data structures.
        """
